from __future__ import division

from scipy.stats import norm, gamma as gammaDist, beta as betaDist
from scipy.special import gammaln
from random import random
from numpy import zeros, array, isnan
import pylab
from cookbook.frange import frange
from math import sqrt, log, exp
import cPickle as pickle

import sys

def normal(m, s):
    return norm.rvs(m, s)

def invGamma(alpha, beta):
    return 1.0 / gammaDist.rvs(alpha, 1/beta)

def bernoulli(p):
    if isnan(p):
        p = .5
    return int(random() < p)

def genData(sigma, a, b, u, v, n=100):
    ll = 0
    #m ~ N(0, sigma) : means of distrs
    ms = []
    for cluster in range(2):
        mu = []
        for dim in range(2):
            mi = normal(0, sigma)
            mu.append(mi)
            ll += log(norm.pdf(mi, 0, sigma))
        mu = array(mu)
        print ">>>", mu
        ms.append(mu)

    #v ~ Ident * InvGamma(a, b) : independent covariance terms
    v1 = zeros((2, 2))
    v1[0][0] = invGamma(a, b)
    v1[1][1] = invGamma(a, b)

    v2 = zeros((2, 2))
    v2[0][0] = invGamma(a, b)
    v2[1][1] = invGamma(a, b)

    ll += log(gammaDist.pdf(1/v1[0][0], a, 1/b))
    ll += log(gammaDist.pdf(1/v1[1][1], a, 1/b))
    ll += log(gammaDist.pdf(1/v2[0][0], a, 1/b))
    ll += log(gammaDist.pdf(1/v2[1][1], a, 1/b))

    #log-likelihood terms, with and without explicit pi
    ll1 = 0
    ll2 = 0

    #pi ~ Beta(u, v) : mixture probability
    pi = betaDist.rvs(u, v)

    ll1 += log(betaDist.pdf(pi, u, v) + 1e-20)

    vs = [v1, v2]

    xs = []
    zs = []
    n0 = 0
    n1 = 0
    for pt in range(n):
        #z ~ Bernoulli(pi) : which cluster?
        z = bernoulli(pi)

        if z == 0:
            ll1 += log(1 - pi)
            n0 += 1
        else:
            ll1 += log(pi)
            n1 += 1

        m = ms[z]
        var = vs[z]
        #x ~ N(m_z, v_z) : 2d datapoint
        x = [normal(m[0], var[0][0]), normal(m[1], var[1][1])]

        ll += log(norm.pdf(x[0], m[0], var[0][0]))
        ll += log(norm.pdf(x[1], m[1], var[1][1]))

        xs.append(x)
        zs.append(z)

    zTerm = gammaln(n0 + u) + gammaln(n1 + v) - gammaln(n0 + n1 + u + v)
    ll2 += zTerm

    ll1 += ll
    ll2 += ll

    return (ms, vs, pi, zs, xs, ll1, ll2)

def plotData(ms, vs, zs, xs):
    pylab.figure()
    for cluster, color in zip([0,1], ['r', 'b']):
        xc = []
        yc = []
        for z,x in zip(zs, xs):
            if z == cluster:
                xc.append(x[0])
                yc.append(x[1])
        pylab.plot(xc, yc, "%s." % color)
    minx = min([p[0] for p in xs])
    miny = min([p[1] for p in xs])
    maxx = max([p[0] for p in xs])
    maxy = max([p[1] for p in xs])

    for cluster, color in zip([0,1], ['r', 'b']):
        m = ms[cluster]
        v = vs[cluster]

        xPts = []
        yPts = []
        zPts = []
        for py in frange(miny, maxy, .1):
            xPts.append([])
            yPts.append([])
            zPts.append([])
            for px in frange(minx, maxx, .1):
                xPts[-1].append(px)
                yPts[-1].append(py)
                xProb = norm.pdf(px, m[0], v[0][0])
                yProb = norm.pdf(py, m[1], v[1][1])
                zPts[-1].append(xProb * yProb)
        lineAt = .6 * (norm.pdf(m[0], m[0], v[0][0]) *
                       norm.pdf(m[1], m[1], v[1][1]))

        pylab.contour(xPts, yPts, zPts, [lineAt,], colors=color)

if __name__ == "__main__":
    #flat beta prior
    u = 10
    v = 10

    if False:
        #these bad hyperparameters tend to create a single cluster
        u = .01
        v = .01

    numPts = 100

    if False:
        #check out inference on too little data! It's terrible!
        numPts = 2
    elif False:
        #more data
        numPts = 1000

    (ms, vs, pi, zs, xs, ll1, ll2) = genData(1, 1, 10, u, v, n=numPts)
    print "Pi", pi
    print ll1
    print ll2
    plotData(ms, vs, zs, xs)
    pylab.show()

    if len(sys.argv) > 1:
        out = file(sys.argv[1], 'wb')
        pickle.dump(ms, out)
        pickle.dump(vs, out)
        pickle.dump(pi, out)
        pickle.dump(zs, out)
        pickle.dump(xs, out)
        pickle.dump(ll1, out)
        pickle.dump(ll2, out)
